import errno
import warnings
from pathlib import Path
from typing import Optional, Tuple

# This import is needed for alembic to recognize the geopackage dialect
import geoalchemy2.alembic_helpers  # noqa: F401
import sqlalchemy as sa
from alembic import command as alembic_command
from alembic.config import Config
from alembic.environment import EnvironmentContext
from alembic.migration import MigrationContext
from alembic.script import ScriptDirectory
from geoalchemy2.admin.dialects.geopackage import create_spatial_ref_sys_view
from geoalchemy2.functions import ST_SRID
from osgeo import gdal, ogr, osr
from sqlalchemy import Column, Integer, MetaData, Table, text

from ..domain import constants, models
from ..infrastructure.spatial_index import ensure_spatial_indexes
from .errors import InvalidSRIDException, MigrationMissingError, UpgradeFailedError
from .upgrade_utils import get_upgrade_steps_count, setup_logging

gdal.UseExceptions()

__all__ = ["ModelSchema"]


def get_alembic_config(engine=None, unsafe=False):
    alembic_cfg = Config()
    alembic_cfg.set_main_option("script_location", "threedi_schema:migrations")
    alembic_cfg.set_main_option("version_table", constants.VERSION_TABLE_NAME)
    if engine is not None:
        alembic_cfg.attributes["engine"] = engine
    alembic_cfg.attributes["unsafe"] = unsafe
    return alembic_cfg


def get_schema_version():
    """Returns the version of the schema in this library"""
    config = get_alembic_config()
    script = ScriptDirectory.from_config(config)
    with EnvironmentContext(config=config, script=script) as env:
        return int(env.get_head_revision())


def _upgrade_database(db, revision="head", unsafe=True):
    """Upgrade ThreediDatabase instance"""
    engine = db.engine
    config = get_alembic_config(engine, unsafe=unsafe)
    alembic_command.upgrade(config, revision)


class GdalErrorHandler:
    def __call__(self, err_level, err_no, err_msg):
        self.err_level = err_level
        self.err_no = err_no
        self.err_msg = err_msg


class ModelSchema:
    def __init__(self, threedi_db, declared_models=models.DECLARED_MODELS):
        self.db = threedi_db
        self.declared_models = declared_models

    def _get_version_old(self):
        """The version of the database using the old 'south' versioning."""
        south_migrationhistory = Table(
            "south_migrationhistory", MetaData(), Column("id", Integer)
        )
        engine = self.db.engine
        if not self.db.has_table("south_migrationhistory"):
            return
        with engine.connect() as connection:
            query = south_migrationhistory.select().order_by(
                south_migrationhistory.columns["id"].desc()
            )
            versions = list(connection.execute(query.limit(1)))
            if len(versions) == 1:
                return versions[0][0]
            else:
                return None

    def get_version(self):
        """Returns the id (integer) of the latest migration"""
        with self.db.engine.connect() as connection:
            context = MigrationContext.configure(
                connection, opts={"version_table": constants.VERSION_TABLE_NAME}
            )
            version = context.get_current_revision()
        if version is not None:
            return int(version)
        else:
            return self._get_version_old()

    def _get_epsg_data(self) -> Tuple[Optional[int], str]:
        """
        Retrieve epsg code for schematisation loaded in session. This is done by
        iterating over all geometries in the declared models and all raster files, and
        stopping at the first geometry or raster file with data.

        Returns the epsg code and the name (table.column) of the source.
        """
        session = self.db.get_session()
        version = self.get_version()

        # for revision < 230 read explicit epsg from schematisation
        if version is not None and version < 230:
            try:
                epsg_code = get_model_srid(
                    connection=session, v2_global_settings=version < 222
                )
            except InvalidSRIDException:
                return None, ""

            return (
                epsg_code,
                "v2_global_settings.epsg_code"
                if version < 222
                else "model_settings.epsg_code",
            )
        # for version 230 (implicit crs in spatialite) get epsg from first geometry object found in the model
        elif version == 230:
            for model in self.declared_models:
                if hasattr(model, "geom"):
                    srids = [
                        item[0] for item in session.query(ST_SRID(model.geom)).all()
                    ]
                    if len(srids) > 0:
                        return srids[0], f"{model.__tablename__}.geom"
            return None, ""
        # for version >= 300 (implicit crs in geopackage) get epsg from connection_node table in geopackage
        else:
            datasource = ogr.Open(str(self.db.path))
            layer = datasource.GetLayerByName("connection_node")
            epsg = layer.GetSpatialRef().GetAuthorityCode(None)
            try:
                epsg = int(epsg)
            except TypeError:
                raise InvalidSRIDException(epsg, "the epsg_code must be an integer")
            return epsg, ""

    def _get_dem_epsg(self, raster_path: str = None) -> int:
        """
        Extract EPSG code from DEM.

        Only works in local filesystem. The raster path references do not resolve correctly in the object store.
        """
        if not raster_path:
            with self.db.get_session() as session:
                settings_table = (
                    "v2_global_settings"
                    if self.get_version() < 222
                    else "model_settings"
                )
                raster_path = session.execute(
                    text(f"SELECT dem_file FROM {settings_table};")
                ).scalar()
            if raster_path is None:
                raise InvalidSRIDException(None, "no DEM is provided")
        # old dem paths include rasters/ but new ones do not
        # to work around this, we remove "rasters/" if present and then add it again
        raster_path = raster_path.replace("\\", "/").split("/")[-1]
        directory = Path(self.db.path).parent
        raster_path = str(directory / "rasters" / Path(raster_path))
        try:
            dataset = gdal.Open(raster_path)
        except RuntimeError as e:
            raise InvalidSRIDException(f"Cannot open filepath {raster_path}") from e
        proj = osr.SpatialReference(wkt=dataset.GetProjection())
        return int(proj.GetAuthorityCode("PROJCS"))

    @property
    def epsg_code(self):
        """
        Raises threedi_schema.migrations.exceptions.InvalidSRIDException if the epsg_code count not be determined or is invalid.
        """
        return self._get_epsg_data()[0]

    @property
    def epsg_source(self):
        """
        Raises threedi_schema.migrations.exceptions.InvalidSRIDException if the epsg_code count not be determined or is invalid.
        """
        return self._get_epsg_data()[1]

    @property
    def is_geopackage(self):
        with self.db.get_session() as session:
            return bool(
                session.execute(
                    text(
                        "SELECT count(*) FROM sqlite_master WHERE type='table' AND name='gpkg_contents';"
                    )
                ).scalar()
            )

    @property
    def is_spatialite(self):
        with self.db.get_session() as session:
            return bool(
                session.execute(
                    text(
                        "SELECT count(*) FROM sqlite_master WHERE type='table' AND name='spatial_ref_sys';"
                    )
                ).scalar()
            )

    def upgrade(
        self,
        revision="head",
        backup=True,
        progress_func=None,
        epsg_code_override=None,
        keep_spatialite=False,
    ):
        """Upgrade the database to the latest version.

        This requires either a completely empty database or a database with its
        current schema version at least 174 (the latest migration of the old
        model databank).

        The upgrade is done using database transactions. However, for SQLite,
        database transactions are only partially supported. To ensure that the
        database file does not become corrupt, enable the "backup" parameter.
        If the database is temporary already (or if it is PostGIS), disable
        it.

        Specify a 'progress_func' to handle progress updates. `progress_func` should
        expect two arguments: the percentage of progress and a string describing the migration step

        Specify a `epsg_code_override` to set the model epsg_code before migration.
        This can be used for testing and for setting the DEM epsg_code when self.epsg_code is None.
        """
        try:
            rev_nr = get_schema_version() if revision == "head" else int(revision)
        except ValueError:
            raise ValueError(
                f"Incorrect version format: {revision}. Expected 'head' or a numeric value."
            )
        v = self.get_version()

        if v is not None and v < constants.LATEST_SOUTH_MIGRATION_ID:
            raise MigrationMissingError(
                f"This tool cannot update versions below "
                f"{constants.LATEST_SOUTH_MIGRATION_ID}. Please consult the "
                f"3Di documentation on how to update legacy databases."
            )
        if (
            v is not None
            and v <= constants.LAST_SPTL_SCHEMA_VERSION
            and not self.is_spatialite
        ):
            raise UpgradeFailedError(
                f"Cannot upgrade from {revision=} because {self.db.path} is not a spatialite"
            )
        elif (
            v is not None
            and v > constants.LAST_SPTL_SCHEMA_VERSION
            and not self.is_geopackage
        ):
            raise UpgradeFailedError(
                f"Cannot upgrade from {revision=} because {self.db.path} is not a geopackage"
            )

        if progress_func is not None:
            config = get_alembic_config(self.db.engine, unsafe=backup)
            n_steps = get_upgrade_steps_count(
                config, self.db.schema.get_version(), revision
            )
            setup_logging(progress_func, n_steps)

        def run_upgrade(_revision):
            if backup:
                with self.db.file_transaction() as work_db:
                    _upgrade_database(work_db, revision=_revision, unsafe=True)
            else:
                _upgrade_database(self.db, revision=_revision, unsafe=False)

        if epsg_code_override is not None:
            if self.get_version() is not None and self.get_version() > 229:
                warnings.warn(
                    "Cannot set epsg_code_override when upgrading from 230 or newer"
                )
            elif rev_nr < 230:
                warnings.warn(
                    "Warning: cannot set epsg_code_override when upgrading to 229 or older."
                )
            else:
                if self.get_version() is None or self.get_version() < 229:
                    run_upgrade("0229")
                self._set_custom_epsg_code(epsg_code_override)
                run_upgrade("0230")
                self._remove_temporary_model_settings()
        # First upgrade to LAST_SPTL_SCHEMA_VERSION.
        # When the requested revision <= LAST_SPTL_SCHEMA_VERSION, this is the only upgrade step
        run_upgrade(
            revision
            if rev_nr <= constants.LAST_SPTL_SCHEMA_VERSION
            else f"{constants.LAST_SPTL_SCHEMA_VERSION:04d}"
        )
        # Finish upgrade if target revision > LAST_SPTL_SCHEMA_VERSION
        if rev_nr > constants.LAST_SPTL_SCHEMA_VERSION:
            self.convert_to_geopackage(delete_spatialite=not keep_spatialite)
            run_upgrade(revision)

    def _set_custom_epsg_code(self, custom_epsg_code: int):
        """Temporarily set epsg code in model settings for migration 230"""
        if (
            self.get_version() is None
            or self.get_version() < 222
            or self.get_version() > 229
        ):
            raise ValueError(f"Cannot set epsg code for revision {self.get_version()}")
        # modify epsg_code
        with self.db.get_session() as session:
            settings_row_count = session.execute(
                text("SELECT COUNT(id) FROM model_settings;")
            ).scalar()
            # to update empty databases, they must have model_settings.epsg_code set
            if settings_row_count == 0:
                session.execute(
                    text(
                        f"INSERT INTO model_settings (id, epsg_code) VALUES (99999, {custom_epsg_code});"
                    )
                )
            else:
                session.execute(
                    text(f"UPDATE model_settings SET epsg_code = {custom_epsg_code};")
                )
            session.commit()

    def _remove_temporary_model_settings(self):
        """Remove temporary model settings entry introduced for the epsg code"""
        with self.db.get_session() as session:
            session.execute(text("DELETE FROM model_settings WHERE id = 99999;"))
            session.commit()

    def validate_schema(self):
        """Very basic validation of 3Di schema.

        Check that the database has the latest migration applied. If the
        latest migrations is applied, we assume the database also contains all
        tables and columns defined in threedi_model.models.py.

        :return: True if the threedi_db schema is valid, raises an error otherwise.
        :raise MigrationMissingError, MigrationTooHighError
        """
        version = self.get_version()
        schema_version = get_schema_version()
        if version is None or version < schema_version:
            raise MigrationMissingError(
                f"This tool requires at least schema version "
                f"{schema_version}. Current version: {version}."
            )

        if version > schema_version:
            warnings.warn(
                f"The database version is higher than the threedi-schema version "
                f"({version} > {schema_version}). This may lead to unexpected "
                f"results. "
            )
        return True

    def set_spatial_indexes(self):
        """(Re)create spatial indexes in the spatialite according to the latest definitions."""
        version = self.get_version()
        schema_version = get_schema_version()
        if version != schema_version:
            raise MigrationMissingError(
                f"Setting views requires schema version "
                f"{schema_version}. Current version: {version}."
            )

        ensure_spatial_indexes(self.db.engine, models.DECLARED_MODELS)

    def convert_to_geopackage(self, delete_spatialite=True):
        """
        Convert spatialite to geopackage using gdal.VectorTranslate.

        Does nothing if the current database is already a geopackage.

        Raises UpgradeFailedError if the conversion of spatialite to geopackage with VectorTranslate fails.
        """

        handler = GdalErrorHandler()
        gdal.PushErrorHandler(handler)
        gdal.UseExceptions()

        warnings_list = []

        if self.is_geopackage:
            return

        # Ensure database is upgraded and views are recreated
        revision = self.get_version()
        if revision is None or revision < constants.LAST_SPTL_SCHEMA_VERSION:
            self.upgrade(
                revision=f"{constants.LAST_SPTL_SCHEMA_VERSION:04d}", backup=False
            )
        elif revision > constants.LAST_SPTL_SCHEMA_VERSION:
            UpgradeFailedError(
                f"Cannot convert schema version {revision} to geopackage"
            )
        # Make necessary modifications for conversion on temporary database
        with self.db.file_transaction(start_empty=False, copy_results=False) as work_db:
            # remove spatialite specific tables that break conversion
            with work_db.get_session() as session:
                session.execute(text("DROP TABLE IF EXISTS spatialite_history;"))
                session.execute(text("DROP TABLE IF EXISTS views_geometry_columns;"))

                all_tablenames = [model.__tablename__ for model in self.declared_models]
                geometry_tablenames = (
                    session.execute(text("SELECT f_table_name FROM geometry_columns;"))
                    .scalars()
                    .all()
                )
                non_geometry_tablenames = list(
                    set(all_tablenames) - set(geometry_tablenames)
                )

                if (
                    session.execute(
                        text(
                            "SELECT count(*) FROM sqlite_master WHERE name='schema_version';"
                        )
                    ).scalar()
                    > 0
                ):
                    non_geometry_tablenames.append("schema_version")

            infile = str(work_db.path)
            outfile = str(Path(self.db.path).with_suffix(".gpkg"))

            conversion_list = []
            conversion_list.append(
                gdal.VectorTranslateOptions(
                    format="gpkg",
                    skipFailures=True,
                )
            )
            for table in non_geometry_tablenames:
                conversion_list.append(
                    gdal.VectorTranslateOptions(
                        format="gpkg",
                        accessMode="update",
                        layers=[table],
                        layerName=table,
                        options=["-preserve_fid"],
                    )
                )
            for conversion_options in conversion_list:
                try:
                    ds = gdal.VectorTranslate(
                        destNameOrDestDS=outfile,
                        srcDS=infile,
                        options=conversion_options,
                    )
                    # dereference dataset before writing additional layers to ensure the data is written
                    del ds
                except RuntimeError as err:
                    raise UpgradeFailedError from err
                else:
                    if (
                        hasattr(handler, "err_level")
                        and handler.err_level >= gdal.CE_Warning
                        and handler.err_msg != "Feature id 0 not preserved"
                    ):
                        warnings_list.append(handler.err_msg)

            if len(warnings_list) > 0:
                warning_string = "\n".join(
                    ["GeoPackage conversion didn't finish as expected:"] + warnings_list
                )
                warnings.warn(warning_string)

                # Correct path of current database
        self.db.path = Path(self.db.path).with_suffix(".gpkg")
        # Reset engine so new path is used on the next call of get_engine()
        self.db._engine = None
        # Recreate views_geometry_columns so set_views works as expected
        with self.db.get_session() as session:
            session.execute(
                text(
                    "CREATE TABLE views_geometry_columns(view_name TEXT, view_geometry TEXT, view_rowid TEXT, f_table_name VARCHAR(256), f_geometry_column VARCHAR(256))"
                )
            )
            create_spatial_ref_sys_view(session)
        ensure_spatial_indexes(self.db.engine, models.DECLARED_MODELS)
        # delete spatialite after none of the steps raised an error
        if delete_spatialite:
            self._delete_spatialite()

    def _delete_spatialite(self):
        """
        Delete spatialite only when the schematisation is a geopackage and handle errors while deleting.
        """
        if self.is_spatialite:
            warnings.warn(
                "Skipped deleting spatialite because this is a spatialite schematisation!"
            )
        if self.is_geopackage:
            sptl_path = Path(self.db.path).with_suffix(".sqlite")
            try:
                # Try to remove the file and do not worry if it was already removed
                sptl_path.unlink(missing_ok=True)
            except PermissionError as e:
                warnings.warn(
                    f"Permission denied removing {sptl_path}: {e}", UserWarning
                )
            except OSError as e:
                if e.errno == errno.EROFS:
                    warnings.warn(
                        "Attempting to remove file on read-only filesystem.",
                        UserWarning,
                    )
                else:
                    warnings.warn(
                        f"Unable to remove old file {sptl_path}: {e}", UserWarning
                    )


def get_model_srid(connection, v2_global_settings: bool = False) -> int:
    table = "v2_global_settings" if v2_global_settings else "model_settings"
    srid_str = connection.execute(sa.text(f"SELECT epsg_code FROM {table}")).fetchone()
    if srid_str is None or srid_str[0] is None:
        raise InvalidSRIDException(None, "no epsg_code is defined")
    try:
        srid = int(srid_str[0])
    except TypeError:
        raise InvalidSRIDException(srid_str[0], "the epsg_code must be an integer")
    return srid



"""List nullable foreign keys and set them to NULL if they refer to a nonexisting object.

Revision ID: 0204
Revises:
Create Date: 2021-09-29 13:50:19.544275

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "0204"
down_revision = "0203"
branch_labels = None
depends_on = None


NULLABLE_FOREIGN_KEYS = [
    {
        "table": "v2_global_settings",
        "column": "interflow_settings_id",
        "referring": "v2_interflow",
    },
    {
        "table": "v2_global_settings",
        "column": "control_group_id",
        "referring": "v2_control_group",
    },
    {
        "table": "v2_global_settings",
        "column": "simple_infiltration_settings_id",
        "referring": "v2_simple_infiltration",
    },
    {
        "table": "v2_global_settings",
        "column": "groundwater_settings_id",
        "referring": "v2_groundwater",
    },
    {
        "table": "v2_aggregation_settings",
        "column": "global_settings_id",
        "referring": "v2_global_settings",
    },
]


def upgrade():
    for fk in NULLABLE_FOREIGN_KEYS:
        upgrade_single(fk)


def upgrade_single(fk):
    """Replace non-existing ForeignKey references with NULL"""

    # construct dummy tables to be able to use the ORM
    table = sa.table(
        fk["table"],
        sa.column("id", sa.Integer),
        sa.column(fk["column"], sa.Integer),
    )
    referring = sa.table(
        fk["referring"],
        sa.column("id", sa.Integer),
    )
    column = getattr(table.c, fk["column"])

    # execute the query
    op.execute(
        table.update()
        .where(column.notin_(sa.orm.Query(referring)) & (column != None))
        .values({fk["column"]: sa.null()})
    )


def downgrade():
    pass



from sqlalchemy import text
from sqlalchemy.exc import OperationalError

__all__ = ["recreate_views"]


def drop_view(connection, name):
    try:
        connection.execute(text(f"SELECT DropTable(NULL, '{name}')"))
    except OperationalError:
        connection.execute(text(f"SELECT DropGeoTable('{name}')"))
        connection.execute(text(f"DROP VIEW IF EXISTS '{name}'"))
    connection.execute(
        text(f"DELETE FROM views_geometry_columns WHERE view_name = '{name}'")
    )


def recreate_views(db, file_version, all_views, views_to_delete):
    """Recreate predefined views in a ThreediDatabase instance"""
    engine = db.engine
    with engine.connect() as connection:
        with connection.begin():
            for name, view in all_views.items():
                drop_view(connection, name)

                connection.execute(text(f"CREATE VIEW {name} AS {view['definition']}"))
                if file_version == 3:
                    connection.execute(
                        text(
                            f"INSERT INTO views_geometry_columns (view_name, view_geometry,view_rowid,f_table_name,f_geometry_column) VALUES('{name}', '{view['view_geometry']}', '{view['view_rowid']}', '{view['f_table_name']}', '{view['f_geometry_column']}')"
                        )
                    )
                else:
                    connection.execute(
                        text(
                            f"INSERT INTO views_geometry_columns (view_name, view_geometry,view_rowid,f_table_name,f_geometry_column,read_only) VALUES('{name}', '{view['view_geometry']}', '{view['view_rowid']}', '{view['f_table_name']}', '{view['f_geometry_column']}', 0)"
                        )
                    )
            for name in views_to_delete:
                drop_view(connection, name)



"""rename vegetation columns

Revision ID: 0217
Revises: 0216
Create Date: 2023-05-05 11:52:34.238859

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = '0217'
down_revision = '0216'
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table("v2_vegetation_drag") as batch_op:
        batch_op.alter_column(
            "height", new_column_name="vegetation_height"
        )
        batch_op.alter_column(
            "height_file", new_column_name="vegetation_height_file"
        )
        batch_op.alter_column(
            "stem_count", new_column_name="vegetation_stem_count"
        )
        batch_op.alter_column(
            "stem_count_file", new_column_name="vegetation_stem_count_file"
        )
        batch_op.alter_column(
            "stem_diameter", new_column_name="vegetation_stem_diameter"
        )
        batch_op.alter_column(
            "stem_diameter_file", new_column_name="vegetation_stem_diameter_file"
        )
        batch_op.alter_column(
            "drag_coefficient", new_column_name="vegetation_drag_coefficient"
        )
        batch_op.alter_column(
            "drag_coefficient_file", new_column_name="vegetation_drag_coefficient_file"
        )


def downgrade():
    with op.batch_alter_table("v2_vegetation_drag") as batch_op:
        batch_op.alter_column(
            "vegetation_height", new_column_name="height"
        )
        batch_op.alter_column(
            "vegetation_height_file", new_column_name="height_file"
        )
        batch_op.alter_column(
            "vegetation_stem_count", new_column_name="stem_count"
        )
        batch_op.alter_column(
            "vegetation_stem_count_file", new_column_name="stem_count_file"
        )
        batch_op.alter_column(
            "vegetation_stem_diameter", new_column_name="stem_diameter"
        )
        batch_op.alter_column(
            "vegetation_stem_diameter_file", new_column_name="stem_diameter_file"
        )
        batch_op.alter_column(
            "vegetation_drag_coefficient", new_column_name="drag_coefficient"
        )
        batch_op.alter_column(
            "vegetation_drag_coefficient_file", new_column_name="drag_coefficient_file"
        )



from pathlib import Path
from unittest import mock

import pytest
from sqlalchemy import Column, inspect, Integer, MetaData, String, Table, text

from threedi_schema import ModelSchema
from threedi_schema.application import errors
from threedi_schema.application.schema import get_schema_version
from threedi_schema.domain import constants
from threedi_schema.domain.models import DECLARED_MODELS
from threedi_schema.infrastructure.spatial_index import get_missing_spatial_indexes


@pytest.fixture
def south_migration_table(in_memory_sqlite):
    south_migrationhistory = Table(
        "south_migrationhistory", MetaData(), Column("id", Integer)
    )
    engine = in_memory_sqlite.engine
    south_migrationhistory.create(engine)
    return south_migrationhistory


@pytest.fixture
def alembic_version_table(in_memory_sqlite):
    alembic_version = Table(
        constants.VERSION_TABLE_NAME,
        MetaData(),
        Column("version_num", String(32), nullable=False),
    )
    engine = in_memory_sqlite.engine
    alembic_version.create(engine)
    return alembic_version


def test_get_schema_version():
    """The current version in the library. We start counting at 200."""
    # this will catch future mistakes of setting non-integer revisions
    assert get_schema_version() >= 200


def test_get_version_no_tables(in_memory_sqlite):
    """Get the version of a sqlite with no version tables"""
    schema_checker = ModelSchema(in_memory_sqlite)
    migration_id = schema_checker.get_version()
    assert migration_id is None


def test_get_version_empty_south(in_memory_sqlite, south_migration_table):
    """Get the version of a sqlite with an empty South version table"""
    schema_checker = ModelSchema(in_memory_sqlite)
    migration_id = schema_checker.get_version()
    assert migration_id is None


def test_get_version_south(in_memory_sqlite, south_migration_table):
    """Get the version of a sqlite with a South version table"""
    with in_memory_sqlite.engine.connect() as connection:
        with connection.begin():
            for v in (42, 43):
                connection.execute(south_migration_table.insert().values(id=v))

    schema_checker = ModelSchema(in_memory_sqlite)
    migration_id = schema_checker.get_version()
    assert migration_id == 43


def test_get_version_empty_alembic(in_memory_sqlite, alembic_version_table):
    """Get the version of a sqlite with an empty alembic version table"""
    schema_checker = ModelSchema(in_memory_sqlite)
    migration_id = schema_checker.get_version()
    assert migration_id is None


def test_get_version_alembic(in_memory_sqlite, alembic_version_table):
    """Get the version of a sqlite with an alembic version table"""
    with in_memory_sqlite.engine.connect() as connection:
        with connection.begin():
            connection.execute(
                alembic_version_table.insert().values(version_num="0201")
            )

    schema_checker = ModelSchema(in_memory_sqlite)
    migration_id = schema_checker.get_version()
    assert migration_id == 201


def test_validate_schema(sqlite_latest):
    """Validate a correct schema version"""
    schema = sqlite_latest.schema
    with mock.patch.object(schema, "get_version", return_value=get_schema_version()):
        assert schema.validate_schema()


@pytest.mark.parametrize("version", [-1, 205, None])
def test_validate_schema_missing_migration(sqlite_latest, version):
    """Validate a too low schema version"""
    schema = ModelSchema(sqlite_latest)
    with mock.patch.object(schema, "get_version", return_value=version):
        with pytest.raises(errors.MigrationMissingError):
            schema.validate_schema()


@pytest.mark.parametrize("version", [9999])
def test_validate_schema_too_high_migration(sqlite_latest, version):
    """Validate a too high schema version"""
    schema = ModelSchema(sqlite_latest)
    with mock.patch.object(schema, "get_version", return_value=version):
        with pytest.warns(UserWarning):
            schema.validate_schema()


def get_sql_tables(cursor):
    return [
        item[0]
        for item in cursor.execute(
            text("SELECT name FROM sqlite_master WHERE type='table';")
        ).fetchall()
    ]


def get_columns_from_schema(schema, table_name):
    inspector = inspect(schema.db.get_engine())
    return [column["name"] for column in inspector.get_columns(table_name)]


def get_columns_from_sqlite(session, table_name):
    res = session.execute(text(f"PRAGMA table_info({table_name})")).fetchall()
    return [col_info[1] for col_info in res]


def test_full_upgrade_oldest(oldest_sqlite):
    """Upgrade a legacy database to the latest version"""
    schema = ModelSchema(oldest_sqlite)
    schema.upgrade(backup=False)
    run_upgrade_test(schema)


def test_full_upgrade_empty(empty_sqlite_v4):
    """Upgrade an empty database to the latest version"""
    schema = ModelSchema(empty_sqlite_v4)
    schema.upgrade(backup=False, epsg_code_override=28992)
    run_upgrade_test(schema)


def test_full_upgrade_with_preexisting_version(south_latest_sqlite):
    """Upgrade an empty database to the latest version"""
    schema = ModelSchema(south_latest_sqlite)
    schema.upgrade(backup=False, epsg_code_override=28992)
    run_upgrade_test(schema)


def run_upgrade_test(schema):
    assert schema.get_version() == get_schema_version()
    session = schema.db.get_session()
    sqlite_tables = get_sql_tables(session)
    schema_tables = [model.__tablename__ for model in schema.declared_models]
    assert set(schema_tables).issubset(set(sqlite_tables))
    for table in schema_tables:
        schema_cols = get_columns_from_schema(schema, table)
        sqlite_cols = get_columns_from_sqlite(session, table)
        assert set(schema_cols).issubset(set(sqlite_cols))
    assert get_missing_spatial_indexes(schema.db.engine, DECLARED_MODELS) == []


def test_upgrade_with_epsg_code_override(in_memory_sqlite):
    """Upgrade an empty database to the latest version and set custom epsg"""
    schema = ModelSchema(in_memory_sqlite)
    schema.upgrade(revision="0230", backup=False, epsg_code_override=28992)
    with schema.db.get_session() as session:
        srids = [
            item[0]
            for item in session.execute(
                text(
                    "SELECT srid FROM geometry_columns WHERE f_table_name NOT LIKE '_%'"
                )
            ).fetchall()
        ]
        assert all([srid == 28992 for srid in srids])


def test_upgrade_with_epsg_code_override_version_too_new(empty_sqlite_v4):
    """Set custom epsg code for schema version > 229"""
    schema = ModelSchema(empty_sqlite_v4)
    schema.upgrade(revision="0230", backup=False, epsg_code_override=28992)
    with pytest.warns():
        schema.upgrade(backup=False, epsg_code_override=28992)


def test_upgrade_with_epsg_code_override_revision_too_old(in_memory_sqlite):
    """Set custom epsg code when upgrading to 228 or older"""
    schema = ModelSchema(in_memory_sqlite)
    with pytest.warns():
        schema.upgrade(revision="0228", backup=False, epsg_code_override=28992)


def test_set_custom_epsg_valid(in_memory_sqlite):
    schema = ModelSchema(in_memory_sqlite)
    schema.upgrade(revision="0229", backup=False)
    schema._set_custom_epsg_code(custom_epsg_code=28992)
    with in_memory_sqlite.engine.connect() as connection:
        check_result = connection.execute(
            text("SELECT epsg_code FROM model_settings")
        ).scalar()
    assert check_result == 28992


@pytest.mark.parametrize(
    "start_revision, epsg_code_override",
    [(None, None), ("0220", None), ("0230", 28992)],
)
def test_set_custom_epsg_invalid_revision(
    in_memory_sqlite, start_revision, epsg_code_override
):
    schema = ModelSchema(in_memory_sqlite)
    if start_revision is not None:
        schema.upgrade(
            revision=start_revision, backup=False, epsg_code_override=epsg_code_override
        )
    with pytest.raises(ValueError):
        schema._set_custom_epsg_code(custom_epsg_code=28992)


def test_upgrade_south_not_latest_errors(in_memory_sqlite):
    """Upgrading a database that is not at the latest south migration will error"""
    schema = ModelSchema(in_memory_sqlite)
    with mock.patch.object(
        schema, "get_version", return_value=constants.LATEST_SOUTH_MIGRATION_ID - 1
    ):
        with pytest.raises(errors.MigrationMissingError):
            schema.upgrade(backup=False)


@pytest.mark.parametrize("keep_spatialite", [True, False])
def test_upgrade_keep_sqlite(south_latest_sqlite, keep_spatialite):
    old_path = Path(south_latest_sqlite.path)
    schema = ModelSchema(south_latest_sqlite)
    schema.upgrade(
        revision="0300", epsg_code_override=28992, keep_spatialite=keep_spatialite
    )
    assert old_path.exists() is keep_spatialite


def test_upgrade_with_backup(south_latest_sqlite):
    """Upgrading with backup=True will proceed on a copy of the database"""
    schema = ModelSchema(south_latest_sqlite)
    with mock.patch(
        "threedi_schema.application.schema._upgrade_database", side_effect=RuntimeError
    ) as upgrade, mock.patch.object(schema, "get_version", return_value=199):
        with pytest.raises(RuntimeError):
            schema.upgrade(backup=True)

    (db,), kwargs = upgrade.call_args
    assert db is not south_latest_sqlite


def test_upgrade_without_backup(south_latest_sqlite):
    """Upgrading with backup=True will proceed on the database itself"""
    schema = ModelSchema(south_latest_sqlite)
    with mock.patch(
        "threedi_schema.application.schema._upgrade_database", side_effect=RuntimeError
    ) as upgrade, mock.patch.object(schema, "get_version", return_value=199):
        with pytest.raises(RuntimeError):
            schema.upgrade(backup=False)

    (db,), kwargs = upgrade.call_args
    assert db is south_latest_sqlite


@pytest.mark.parametrize(
    "is_var, version",
    [("is_spatialite", constants.LAST_SPTL_SCHEMA_VERSION), ("is_geopackage", 300)],
)
def test_upgrade_incorrect_format(in_memory_sqlite, is_var, version):
    schema = ModelSchema(in_memory_sqlite)
    with mock.patch.object(
        type(schema), is_var, new=mock.PropertyMock(return_value=False)
    ):
        with mock.patch.object(type(schema), "get_version", return_value=version):
            with pytest.raises(errors.UpgradeFailedError):
                schema.upgrade()


def test_upgrade_revision_exception(oldest_sqlite):
    schema = ModelSchema(oldest_sqlite)
    with pytest.raises(ValueError):
        schema.upgrade(revision="foo")


# TODO: remove this because ensure_spatial_indexes is already tested
@pytest.mark.skip(reason="will be removed")
def test_set_spatial_indexes(in_memory_sqlite):
    engine = in_memory_sqlite.engine

    schema = ModelSchema(in_memory_sqlite)
    schema.upgrade(backup=False, epsg_code_override=28992)

    with engine.connect() as connection:
        with connection.begin():
            connection.execute(
                text("SELECT DisableSpatialIndex('connection_node', 'geom')")
            ).scalar()
            connection.execute(text("DROP TABLE idx_connection_node_geom"))

    schema.set_spatial_indexes()

    with engine.connect() as connection:
        check_result = connection.execute(
            text("SELECT CheckSpatialIndex('connection_node', 'geom')")
        ).scalar()

    assert check_result == 1


def test_is_spatialite(in_memory_sqlite):
    schema = ModelSchema(in_memory_sqlite)
    schema.upgrade(
        revision=f"{constants.LAST_SPTL_SCHEMA_VERSION:04d}",
        backup=False,
        epsg_code_override=28992,
    )
    assert schema.is_spatialite


def test_is_geopackage(oldest_sqlite):
    schema = ModelSchema(oldest_sqlite)
    schema.upgrade(
        revision=f"{constants.LAST_SPTL_SCHEMA_VERSION:04d}",
        backup=False,
        epsg_code_override=28992,
    )
    schema.convert_to_geopackage()
    assert schema.is_geopackage


@pytest.mark.filterwarnings("ignore::UserWarning")
@pytest.mark.parametrize(
    "revision, expected_epsg_code", [("0229", None), ("0230", None), ("head", 28992)]
)
def test_get_epsg_data_empty(empty_sqlite_v4, revision, expected_epsg_code):
    schema = ModelSchema(empty_sqlite_v4)
    schema.upgrade(revision=revision, backup=False, epsg_code_override=28992)
    if expected_epsg_code is None:
        assert schema.epsg_code is None
    else:
        assert schema.epsg_code == expected_epsg_code
    assert schema.epsg_source == ""


@pytest.mark.parametrize(
    "revision, expected_epsg_source",
    [
        ("0220", "v2_global_settings.epsg_code"),
        ("0229", "model_settings.epsg_code"),
        ("0230", "boundary_condition_1d.geom"),
        ("head", ""),
    ],
)
def test_get_epsg_data(oldest_sqlite, revision, expected_epsg_source):
    schema = ModelSchema(oldest_sqlite)
    schema.upgrade(revision=revision, backup=False)
    assert schema.epsg_code == 28992
    assert schema.epsg_source == expected_epsg_source


def test_epsg_code_from_dem(sqlite_with_dem):
    schema = ModelSchema(sqlite_with_dem)
    assert schema._get_dem_epsg() == 28991
    schema.upgrade(epsg_code_override=schema._get_dem_epsg())
    assert schema._get_dem_epsg() == 28991



from pathlib import Path

import pytest
from sqlalchemy import text

from threedi_schema.domain import constants


@pytest.mark.parametrize("delete_spatialite", [True, False])
def test_convert_to_geopackage(oldest_sqlite, delete_spatialite):
    oldest_sqlite.schema.upgrade(revision=f"{constants.LAST_SPTL_SCHEMA_VERSION:04d}")
    old_path = Path(oldest_sqlite.path)
    oldest_sqlite.schema.convert_to_geopackage(delete_spatialite=delete_spatialite)
    # Ensure that after the conversion the geopackage is used
    assert oldest_sqlite.path.suffix == ".gpkg"
    if delete_spatialite:
        assert not old_path.exists()
    assert not oldest_sqlite.schema.is_spatialite
    assert oldest_sqlite.schema.is_geopackage

    with oldest_sqlite.get_session() as session:
        assert (
            session.execute(
                text(
                    "SELECT count(*) FROM sqlite_master WHERE type='view' AND name='spatial_ref_sys'"
                )
            ).scalar()
            == 1
        )



import pytest
from sqlalchemy import Column, create_engine, func, Integer
from sqlalchemy.event import listen
from sqlalchemy.orm import declarative_base, sessionmaker

from threedi_schema.application.threedi_database import load_spatialite
from threedi_schema.domain.custom_types import (
    clean_csv_string,
    clean_csv_table,
    Geometry,
)


@pytest.mark.parametrize(
    "value",
    [
        "1,2,3",
        "1, 2, 3 ",
        "1,\t2,3",
        "1,\r2,3 ",
        "1,\n2,3 ",
        "1,  2,3",
        "1,  2  ,3",
        " 1,2,3 ",
        "\n1,2,3",
        "\t1,2,3",
        "\r1,2,3",
        "1,2,3\t",
        "1,2,3\n",
        "1,2,3\r",
    ],
)
def test_clean_csv_string(value):
    assert clean_csv_string(value) == "1,2,3"


def test_clean_csv_string_with_whitespace():
    assert clean_csv_string("1,2 3,4") == "1,2 3,4"


@pytest.mark.parametrize(
    "value",
    [
        "1,2,3\n4,5,6",
        "1,2,3\r\n4,5,6",
        "\n1,2,3\n4,5,6",
        "1,2,3\n4,5,6\n",
    ],
)
def test_clean_csv_table(value):
    assert clean_csv_table(value) == "1,2,3\n4,5,6"


@pytest.mark.parametrize(
    "value", [" ", "0 1", "3;5", "foo", "1,2\n3,", ",2", ",2\n3,4"]
)
def test_clean_csv_table_no_fail(value):
    clean_csv_table(value)


@pytest.mark.parametrize("from_text", [None, "GeomFromEWKB"])
def test_geometry_type_from_text(from_text):
    # Esnure that the "from_text" values that are use to
    # create a Geometry do not break downstream
    Base = declarative_base()

    engine = create_engine("sqlite:///:memory:")
    listen(engine, "connect", load_spatialite)

    Base.metadata.create_all(engine)
    Session = sessionmaker(bind=engine)
    session = Session()
    session.execute(func.gpkgCreateBaseTables())

    class Location(Base):
        __tablename__ = "locations"
        id = Column(Integer, primary_key=True)
        geom = Column(Geometry("POINT", from_text=from_text))  # Invalid from_text

    Base.metadata.create_all(engine)

    try:
        # Create a test point using WKT format
        test_location = Location()
        session.add(test_location)
        session.commit()
    finally:
        session.close()



import logging
from typing import Callable, TYPE_CHECKING

from alembic.config import Config
from alembic.script import ScriptDirectory

if TYPE_CHECKING:
    from .schema import ModelSchema
else:
    ModelSchema = None


class ProgressHandler(logging.Handler):
    def __init__(self, progress_func, total_steps):
        super().__init__()
        self.progress_func = progress_func
        self.total_steps = total_steps
        self.current_step = 0

    def emit(self, record):
        msg = record.getMessage()
        if msg.startswith("Running upgrade"):
            if self.total_steps > 0:
                self.progress_func(100 * self.current_step / self.total_steps, msg)
            else:
                self.progress_func(100, msg)  # Assume 100% if total steps are zero
            self.current_step += 1


def get_upgrade_steps_count(
    config: Config, current_revision: int, target_revision: str = "head"
) -> int:
    """
    Count number of upgrade steps for a schematisation upgrade.

    Args:
        config: Config parameter containing the configuration information
        current_revision: current revision as integer
        target_revision: target revision as zero-padded 4 digit string or "head"
    """
    if target_revision != "head":
        try:
            int(target_revision)
        except TypeError:
            # this should lead to issues in the upgrade pipeline, lets not take over that error handling here
            return 0
    # walk_revisions also includes the revision from current_revision to previous
    # reduce the number of steps with 1
    offset = -1
    # The first defined revision is 200; revision numbers < 200 will cause walk_revisions to fail
    if current_revision < 200:
        current_revision = 200
        # set offset to 0 because previous to current is not included in walk_revisions
        offset = 0
    if target_revision != "head" and int(target_revision) < current_revision:
        # assume that this will be correctly handled by alembic
        return 0
    current_revision_str = f"{current_revision:04d}"
    script = ScriptDirectory.from_config(config)
    # Determine upgrade steps
    revisions = script.walk_revisions(current_revision_str, target_revision)
    return len(list(revisions)) + offset


def setup_logging(progress_func: Callable[[float, str], None], n_steps: int):
    """
    Set up logging for schematisation upgrade

    Args:
        schema: ModelSchema object representing the current schema of the application
        target_revision: A str specifying the target revision for migration
        config: Config object containing configuration settings
        progress_func: A Callable with a single argument of type float, used to track progress during migration
    """
    logger = logging.getLogger("alembic.runtime.migration")
    logger.setLevel(logging.INFO)
    handler = ProgressHandler(progress_func, total_steps=n_steps)
    logger.addHandler(handler)
    return handler



"""add discharge capacity for orifice

Revision ID: 0301
Revises: 0300
Create Date: 2026-03-12

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "0301"
down_revision = "0300"
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table("orifice") as batch_op:
        batch_op.add_column(sa.Column("discharge_capacity", sa.Float(), nullable=True))


def downgrade():
    with op.batch_alter_table("orifice") as batch_op:
        batch_op.drop_column("discharge_capacity")




"""Convert to geopackage

Revision ID: 0230
Revises:
Create Date: 2024-11-12 12:30

"""
import sqlite3
import uuid

import sqlalchemy as sa
from alembic import op

from threedi_schema.application.errors import InvalidSRIDException

# revision identifiers, used by Alembic.
revision = "0300"
down_revision = "0230"
branch_labels = None
depends_on = None


def fix_use_0d_inflow():
    # fix setting use_0d_inflow to be only 0 or 1 (also see migration 223)
    op.execute(sa.text("""UPDATE simulation_template_settings SET use_0d_inflow = 1 WHERE use_0d_inflow > 0 """))


def upgrade():
    fix_use_0d_inflow()


def downgrade():
    # Not implemented on purpose
    pass



"""Clean up some issues from migrations 222 to 228

Revision ID: 022
9Revises:
Create Date: 2024-11-15 14:18

"""
import uuid
from typing import List

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "0229"
down_revision = "0228"
branch_labels = None
depends_on = None


def get_geom_type(table_name, geo_col_name):
    connection = op.get_bind()
    columns = connection.execute(sa.text(f"PRAGMA table_info('{table_name}')")).fetchall()
    for col in columns:
        if col[1] == geo_col_name:
            return col[2]

def change_types_in_settings_table():
    temp_table_name = f'_temp_229_{uuid.uuid4().hex}'
    table_name = 'model_settings'
    change_types = {'use_d2_rain': 'bool', 'friction_averaging': 'bool'}
    connection = op.get_bind()
    columns = connection.execute(sa.text(f"PRAGMA table_info('{table_name}')")).fetchall()
    # get all column names and types
    skip_cols = ['id', 'the_geom']
    col_names = [col[1] for col in columns if col[1] not in skip_cols]
    old_col_types = [col[2] for col in columns if col[1] not in skip_cols]
    col_types = [change_types.get(col_name, col_type) for col_name, col_type in zip(col_names, old_col_types)]
    # Create new table, insert data, drop original and rename temp to table_name
    col_str = ','.join(['id INTEGER PRIMARY KEY NOT NULL'] + [f'{cname} {ctype}' for cname, ctype in
                                                                  zip(col_names, col_types)])
    op.execute(sa.text(f"CREATE TABLE {temp_table_name} ({col_str});"))
    # Copy data
    op.execute(sa.text(f"INSERT INTO {temp_table_name} (id, {','.join(col_names)}) "
                       f"SELECT id, {','.join(col_names)} FROM {table_name}"))
    op.execute(sa.text(f"DROP TABLE {table_name}"))
    op.execute(sa.text(f"ALTER TABLE {temp_table_name} RENAME TO {table_name};"))


def remove_tables(tables: List[str]):
    for table in tables:
        op.drop_table(table)


def find_tables_by_pattern(pattern: str) -> List[str]:
    connection = op.get_bind()
    query = connection.execute(
        sa.text(f"select name from sqlite_master where type = 'table' and name like '{pattern}'"))
    return [item[0] for item in query.fetchall()]


def remove_old_tables():
    remaining_v2_idx_tables = find_tables_by_pattern('idx_v2_%_the_geom')
    remaining_alembic = find_tables_by_pattern('%_alembic_%_the_geom')
    remove_tables(set(remaining_v2_idx_tables + remaining_alembic))


def clean_geometry_columns():
    """ Remove columns referencing v2 in geometry_columns """
    op.execute(sa.text("""
            DELETE FROM geometry_columns WHERE f_table_name IN (
                SELECT g.f_table_name FROM geometry_columns g
                LEFT JOIN sqlite_master m ON g.f_table_name = m.name
                WHERE m.name IS NULL AND g.f_table_name like "%v2%"
            );
        """))


def clean_by_type(type: str):
    connection = op.get_bind()
    items = [item[0] for item in connection.execute(
        sa.text(f"SELECT tbl_name FROM sqlite_master WHERE type='{type}' AND tbl_name LIKE '%v2%';")).fetchall()]
    for item in items:
        op.execute(f"DROP {type} IF EXISTS {item};")


def update_use_settings():
    # Ensure that use_* settings are only True when there is actual data for them
    use_settings = [
        ('use_groundwater_storage', 'groundwater'),
        ('use_groundwater_flow', 'groundwater'),
        ('use_interflow', 'interflow'),
        ('use_simple_infiltration', 'simple_infiltration'),
        ('use_vegetation_drag_2d', 'vegetation_drag_2d'),
        ('use_interception', 'interception')
    ]
    connection = op.get_bind()  # Get the connection for raw SQL execution
    for setting, table in use_settings:
        use_row = connection.execute(sa.text(f"SELECT {setting} FROM model_settings")).scalar()
        if not use_row:
            continue
        row = connection.execute(sa.text(f"SELECT * FROM {table}")).first()
        use_row = (row is not None)
        if use_row:
            use_row = not all(item in (None, "") for item in row[1:])
        if not use_row:
            connection.execute(sa.text(f"UPDATE model_settings SET {setting} = 0"))

            
def upgrade():
    remove_old_tables()
    clean_geometry_columns()
    clean_by_type('trigger')
    clean_by_type('view')
    update_use_settings()
    change_types_in_settings_table()


def downgrade():
    pass



import pytest
from geoalchemy2 import Geometry
from sqlalchemy import Column, create_engine, func, Integer
from sqlalchemy.event import listen
from sqlalchemy.orm import declarative_base, sessionmaker

from threedi_schema.application.threedi_database import load_spatialite
from threedi_schema.infrastructure.spatial_index import (
    create_spatial_index,
    ensure_spatial_indexes,
    get_missing_spatial_indexes,
)

Base = declarative_base()


class Model(Base):
    __tablename__ = "model"

    id = Column(Integer, primary_key=True)
    geom = Column(Geometry("POINT"))


@pytest.fixture()
def engine():
    engine = create_engine("sqlite:///:memory:")
    listen(engine, "connect", load_spatialite)

    Base.metadata.create_all(engine)
    Session = sessionmaker(bind=engine)
    session = Session()
    session.execute(func.gpkgCreateBaseTables())
    return engine


def test_get_missing_spatial_indexes(engine):
    assert get_missing_spatial_indexes(engine, [Model]) == [Model]


def test_create_spatial_index(engine):
    with engine.connect() as connection:
        with connection.begin():
            create_spatial_index(connection, Model.__table__.columns["geom"])
    assert get_missing_spatial_indexes(engine, [Model]) == []


def test_create_spatial_index_fail(engine):
    with engine.connect() as connection:
        with connection.begin():
            create_spatial_index(connection, Model.__table__.columns["geom"])
            with pytest.raises(RuntimeError):
                create_spatial_index(connection, Model.__table__.columns["geom"])


def test_ensure_spatial_index(engine):
    assert get_missing_spatial_indexes(engine, [Model]) == [Model]
    ensure_spatial_indexes(engine, [Model])
    assert get_missing_spatial_indexes(engine, [Model]) == []



import sqlite3
from typing import List

import sqlalchemy as sa

from threedi_schema.application.errors import InvalidSRIDException


def drop_geo_table(op, table_name: str):
    """

    Safely drop table, taking into account geometry columns

    Parameters:
    op : object
        An object representing the database operation.
    table_name : str
        The name of the table to be dropped.
    """
    op.execute(sa.text(f"SELECT DropTable(NULL, '{table_name}');"))


def drop_conflicting(op, new_tables: List[str]):
    """
    Drop tables from database that conflict with new tables

    Parameters:
    op: The SQLAlchemy operation context to interact with the database.
    new_tables: A list of new table names to be checked for conflicts with existing tables.
    """
    connection = op.get_bind()
    existing_tables = [item[0] for item in connection.execute(
        sa.text("SELECT name FROM sqlite_master WHERE type='table';")).fetchall()]
    for table_name in set(existing_tables).intersection(new_tables):
        drop_geo_table(op, table_name)


def get_crs_info(srid):
    # Create temporary spatialite to find crs unit and projection
    conn = sqlite3.connect(":memory:")
    conn.enable_load_extension(True)
    conn.load_extension("mod_spatialite")
    # Initialite spatialite without any meta data
    conn.execute("SELECT InitSpatialMetaData(1, 'NONE');")
    # Add CRS
    success = conn.execute(f"SELECT InsertEpsgSrid({srid})").fetchone()[0]
    if not success:
        raise InvalidSRIDException(srid, "the supplied epsg_code is invalid")
    # retrieve units and is_projected
    unit = conn.execute(f'SELECT SridGetUnit({srid})').fetchone()[0]
    is_projected = conn.execute(f'SELECT SridIsProjected({srid})').fetchone()[0]
    return unit, is_projected





"""make settings nullable

Revision ID: 0205
Revises: 0204
Create Date: 2021-11-15 16:41:43.316599

"""
from alembic import op

from sqlalchemy import text


# revision identifiers, used by Alembic.
revision = "0205"
down_revision = "0204"
branch_labels = None
depends_on = None


MIGRATION_QUERIES = """
UPDATE v2_global_settings SET frict_type = 2 WHERE frict_type IS NULL;

UPDATE v2_global_settings SET use_0d_inflow = 0 WHERE use_0d_inflow IS NULL;

UPDATE v2_global_settings SET minimum_sim_time_step = min(0.01, sim_time_step)
  WHERE minimum_sim_time_step IS NULL
    OR minimum_sim_time_step <= 0
    OR minimum_sim_time_step > sim_time_step;

UPDATE v2_global_settings SET maximum_sim_time_step = sim_time_step
  WHERE (maximum_sim_time_step IS NULL OR maximum_sim_time_step < sim_time_step);

UPDATE v2_global_settings SET output_time_step = max(1, sim_time_step)
  WHERE output_time_step IS NULL OR output_time_step < sim_time_step;

UPDATE v2_global_settings SET flooding_threshold = 1e-6
  WHERE flooding_threshold IS NULL OR flooding_threshold < 0;

UPDATE v2_global_settings SET frict_avg = 1 WHERE frict_avg NOT IN (0, 1);

UPDATE v2_global_settings SET max_angle_1d_advection = 1.256637 WHERE max_angle_1d_advection = 90;

UPDATE v2_global_settings SET dem_obstacle_detection = 0 WHERE dem_obstacle_detection = '';

UPDATE v2_global_settings SET table_step_size_1d = NULL WHERE table_step_size_1d <= 0;

UPDATE v2_global_settings SET table_step_size_volume_2d = NULL WHERE table_step_size_volume_2d <= 0;

UPDATE v2_interflow SET interflow_type = 0
  WHERE (interflow_type != 0)
    AND ((porosity IS NULL AND (porosity_file IS NULL OR porosity_file == ''))
      OR (impervious_layer_elevation IS NULL)
      OR (hydraulic_conductivity IS NULL AND (hydraulic_conductivity_file IS NULL OR hydraulic_conductivity_file == '')));

UPDATE v2_simple_infiltration SET infiltration_rate = 0 WHERE infiltration_rate < 0;

UPDATE v2_numerical_settings SET precon_cg = 1 WHERE precon_cg > 1;

UPDATE v2_numerical_settings SET cfl_strictness_factor_1d = NULL WHERE cfl_strictness_factor_1d = 0;

UPDATE v2_numerical_settings SET convergence_eps = 1e-7 WHERE convergence_eps = 0;

UPDATE v2_numerical_settings SET convergence_cg = NULL WHERE convergence_cg = 0;

UPDATE v2_numerical_settings SET flow_direction_threshold = NULL WHERE flow_direction_threshold = 0;

UPDATE v2_numerical_settings SET general_numerical_threshold = NULL WHERE general_numerical_threshold = 0;

UPDATE v2_numerical_settings SET minimum_surface_area = NULL WHERE minimum_surface_area = 0;

UPDATE v2_numerical_settings SET limiter_slope_friction_2d = 1 WHERE limiter_slope_friction_2d = 3;

UPDATE v2_numerical_settings SET integration_method = NULL WHERE integration_method = 1;

DELETE FROM v2_aggregation_settings WHERE aggregation_method = 'med' OR aggregation_method IS NULL;

UPDATE v2_aggregation_settings SET flow_variable = 'discharge'
  WHERE (flow_variable = 'discharge_negative' AND aggregation_method = 'cum_negative')
    OR (flow_variable = 'discharge_positive' AND aggregation_method = 'cum_positive');
"""


def upgrade():
    for q in MIGRATION_QUERIES.split(";"):
        op.execute(text(q))

    with op.batch_alter_table("v2_global_settings") as batch_op:
        batch_op.alter_column("output_time_step", nullable=False)
        batch_op.alter_column("nr_timesteps", nullable=True)
        batch_op.alter_column("start_date", nullable=True)
        batch_op.alter_column("frict_type", nullable=False)
        batch_op.alter_column("minimum_sim_time_step", nullable=False)
        batch_op.alter_column("frict_avg", nullable=False)
        batch_op.alter_column("use_0d_inflow", nullable=False)

    with op.batch_alter_table("v2_numerical_settings") as batch_op:
        batch_op.alter_column("max_degree", nullable=True)
        batch_op.alter_column("use_of_cg", nullable=True)
        batch_op.alter_column("use_of_nested_newton", nullable=True)

    with op.batch_alter_table("v2_aggregation_settings") as batch_op:
        batch_op.alter_column("aggregation_method", nullable=False)


def downgrade():
    pass



import re

import geoalchemy2
from packaging import version
from sqlalchemy.types import Integer, Text, TypeDecorator, VARCHAR


class Geometry(geoalchemy2.types.Geometry):
    cache_ok = False

    def __init__(self, geometry_type, from_text="GeomFromEWKT"):
        kwargs = {
            "geometry_type": geometry_type,
            "srid": 4326,
            "spatial_index": True,
            "from_text": from_text,
        }
        if version.parse(geoalchemy2.__version__) < version.parse("0.13.0"):
            kwargs["management"] = True
        super().__init__(**kwargs)


class CustomEnum(TypeDecorator):
    """Column type for storing integer python enums.

    The CustomEnum will make use of the backend 'impl' datatype to store
    the data. Data is converted back it's enum_class python object if possible.
    If not, the raw 'impl' is returned.

    CustomEnum should not be used directly but inherited by a class which
    provides the `impl` datatype.
    """

    def __init__(self, enum_class):
        """

        :param enum_class: instance of enum.Enum
        """
        super().__init__()
        self.enum_class = enum_class
        self.enums = [e.value for e in enum_class]

    def process_bind_param(self, value, dialect):
        if isinstance(value, self.enum_class):
            return value.value
        else:
            return value

    def process_result_value(self, value, dialect):
        if value in self.enums:
            return self.enum_class(value)
        else:
            return value


class IntegerEnum(CustomEnum):
    """Column type for storing integer python enums.

    The CustomEnum will make use of the backend 'impl' datatype to store
    the data. Data is converted back it's enum_class python object if possible.
    If not, the raw 'impl' is returned.

    CustomEnum should not be used directly but inherited by a class which
    provides the `impl` datatype.
    """

    cache_ok = True
    impl = Integer


def clean_csv_string(value: str) -> str:
    return re.sub(r"\s*,\s*", ",", value.strip())


class CSVText(TypeDecorator):
    impl = Text
    cache_ok = True

    def process_bind_param(self, value, dialect):
        if value is not None:
            value = clean_csv_string(value)
        return value

    def process_result_value(self, value, dialect):
        if value is not None:
            value = clean_csv_string(value)
        return value


def clean_csv_table(value: str) -> str:
    # convert windows line endings to unix first
    value = value.replace("\r\n", "\n")
    # remove leading and trailing whitespace
    value = value.strip()
    # clean up each line
    return "\n".join([clean_csv_string(line) for line in value.split("\n")])


class CSVTable(TypeDecorator):
    impl = Text
    cache_ok = True

    def process_bind_param(self, value, dialect):
        if value is not None:
            value = clean_csv_table(value)
        return value

    def process_result_value(self, value, dialect):
        if value is not None:
            value = clean_csv_table(value)
        return value


class VarcharEnum(CustomEnum):
    cache_ok = True
    impl = VARCHAR
